@@ -8,12 +8,12 @@ __PACKAGE__->follow_best_practice;
 __PACKAGE__->mk_accessors(qw/
     tags allowed compiled plain strict_attributes close_open_tags error
     tree escapes direct_attribute params url_finder text_processor linebreaks
-    smileys /);
+    smileys attribute_parser strip_linebreaks attribute_quote /);
 #use Data::Dumper;
 use Carp;
 my $scalar_util = eval "require Scalar::Util; 1";
 
-our $VERSION = '0.13';
+our $VERSION = '0.14';
 
 my %defaults = (
     strict_attributes   => 1,
@@ -21,6 +21,8 @@ my %defaults = (
     linebreaks          => 1,
     smileys             => 0,
     url_finder          => 0,
+    strip_linebreaks    => 1,
+    attribute_quote     => '"',
 );
 sub new {
     my ($class, $args) = @_;
@@ -249,10 +251,11 @@ sub _compile_def {
             # tag attribute or content
             else {
                 my ($escapes, $type) = @$c;
+                my @escapes = @$escapes;
                 my $var = '';
                 my $attributes = $tag->get_attr;
                 if ($type eq 'attr' and @$attributes > 1) {
-                    my $name = shift @$escapes;
+                    my $name = shift @escapes;
                     for my $item (@$attributes[1 .. $#$attributes]) {
                         if ($item->[0] eq $name) {
                             $var = $item->[1];
@@ -273,7 +276,7 @@ sub _compile_def {
                     }
                     $var = $string;
                 }
-                for my $e (@$escapes) {
+                for my $e (@escapes) {
                     my $sub = $esc->{$e};
                     if ($sub) {
                         $var = $sub->($self, $c, $var);
@@ -308,6 +311,7 @@ sub _render_text {
 
 sub parse {
     my ($self, $text, $params) = @_;
+    my $parse_attributes = $self->get_attribute_parser ? $self->get_attribute_parser : $self->can('parse_attributes');
     $self->set_error(undef);
     my $defs = $self->get_tags;
     my $tags = $self->get_allowed || [keys %$defs];
@@ -355,7 +359,7 @@ sub parse {
                 #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$o], ['o']);
                 if ($self->get_close_open_tags) {
                     # we close the tag for you
-                    $self->_finish_tag($o, '[/' . $o->get_name . ']');
+                    $self->_finish_tag($o, '[/' . $o->get_name . ']', 1);
                     $callback_found_tag->($o);
                     $callback_found_tag->($tag);
                 }
@@ -367,10 +371,52 @@ sub parse {
                     $callback_found_tag->($tag);
                 }
             }
+            elsif (ref $tag) {
+                my $def = $defs->{lc $tag->get_name};
+                my $parse = $def->{parse};
+                if ($parse) {
+                    $o->add_content($tag);
+                }
+                else {
+                    my $content = $tag->get_content;
+                    my $string = '';
+                    for my $c (@$content) {
+                        if (ref $c) {
+                            $string .= $c->raw_text( auto_close => 0 );
+                        }
+                        else {
+                            $string .= $c;
+                        }
+                    }
+                    $tag->set_content([$string]);
+                    $o->add_content($tag);
+                }
+            }
             else {
                 $o->add_content($tag);
             }
         }
+        elsif (ref $tag) {
+            my $def = $defs->{lc $tag->get_name};
+            my $parse = $def->{parse};
+            if ($parse) {
+                push @tags, $tag;
+            }
+            else {
+                my $content = $tag->get_content;
+                my $string = '';
+                for my $c (@$content) {
+                    if (ref $c) {
+                        $string .= $c->raw_text( auto_close => 0 );
+                    }
+                    else {
+                        $string .= $c;
+                    }
+                }
+                $tag->set_content([$string]);
+                push @tags, $tag;
+            }
+        }
         else {
             push @tags, $tag;
         }
@@ -434,8 +480,21 @@ sub parse {
                         unshift @not_close, $try;
                     }
                     else {
-                        # unbalanced, just add unparsed text
-                        $callback_found_tag->($_) for $try->_reduce;
+                        # unbalanced
+                        $self->_add_error('unclosed', $try);
+                        if ($self->get_close_open_tags) {
+                            # close
+                            $f = $try;
+                            unshift @not_close, $try;
+                            if (@opened) {
+                                $opened[-1]->add_content('');
+                            }
+                            $self->_finish_tag($try, '[/'. $try->get_name() .']', 1);
+                        }
+                        else {
+                            # just add unparsed text
+                            $callback_found_tag->($_) for $try->_reduce;
+                        }
                     }
                 }
                 if (@not_close) {
@@ -462,14 +521,16 @@ sub parse {
         }
 
         if (defined $tag1) {
+            $in_url = grep { $_->get_class eq 'url' } @opened;
             # short tag
 #            $callback_found_text->($before) if length $before;
             if ($after =~ s{ :// ([^\[]+) \] }{}x) {
                 my $content = $1;
                 my ($attr, $title) = split /\|/, $content, 2;
-                my $tag = Parse::BBCode::Tag->new({
+                my $tag = $self->new_tag({
                         name    => lc $tag1,
                         attr    => [[$attr]],
+                        attr_raw => $attr,
                         content => [(defined $title and length $title) ? $title : ()],
                         start   => "[$tag1://$content]",
                         close   => 0,
@@ -499,19 +560,20 @@ sub parse {
         if ($after) {
             # found start of a tag
             #warn __PACKAGE__.':'.__LINE__.": find attribute for $tag\n";
-            if (
-                ($self->get_direct_attribute and $after =~ s/^(=[^\]]*)?]//)
-                    or
-                ($after =~ s/^( [^\]]*)?\]//)
-            ) {
-                my $attr = $1;
+            my ($ok, $attributes, $attr_string, $end) = $self->$parse_attributes(
+                text => \$after,
+                tag => lc $tag,
+            );
+            if ($ok) {
+                my $attr = $attr_string;
                 $attr = '' unless defined $attr;
                 #warn __PACKAGE__.':'.__LINE__.": found attribute for $tag: $attr\n";
                 my $close = $defs->{lc $tag}->{close};
                 my $def = $defs->{lc $tag};
-                my $open = Parse::BBCode::Tag->new({
+                my $open = $self->new_tag({
                         name    => lc $tag,
-                        attr    => [],
+                        attr    => $attributes,
+                        attr_raw => $attr_string,
                         content => [],
                         start   => "[$tag$attr]",
                         close   => $close,
@@ -520,9 +582,9 @@ sub parse {
                         in_url  => $in_url,
                         type    => 'classic',
                     });
-                my $success = $self->_validate_attr($open, $attr);
+                my $success = 1;
                 my $nested_url = $in_url && $open->get_class eq 'url';
-                if ($success) {
+                {
                     my $last = $opened[-1];
                     if ($last and not $last->get_close and not $close) {
                         $self->_finish_tag($last, '');
@@ -531,11 +593,11 @@ sub parse {
                         $callback_found_tag->($last);
                     }
                 }
-                if ($success && $open->get_single && !$nested_url) {
+                if ($open->get_single && !$nested_url) {
                     $self->_finish_tag($open, '');
                     $callback_found_tag->($open);
                 }
-                elsif ($success && !$nested_url) {
+                elsif (!$nested_url) {
                     push @opened, $open;
                     my $def = $defs->{lc $tag};
                     #warn __PACKAGE__.':'.__LINE__.": $tag $def\n";
@@ -548,7 +610,7 @@ sub parse {
                     else {
                         #warn __PACKAGE__.':'.__LINE__.": noparse, find content\n";
                         # just search for closing tag
-                        if ($after =~ s# (.*?) (\[ / $tag \]) ##xs) {
+                        if ($after =~ s# (.*?) (\[ / $tag \]) ##ixs) {
                             my $content = $1;
                             my $end = $2;
                             #warn __PACKAGE__.':'.__LINE__.": CONTENT $content\n";
@@ -569,7 +631,7 @@ sub parse {
             }
             else {
                 # unclosed tag
-                $callback_found_text->("[$tag");
+                $callback_found_text->("[$tag$attr_string$end");
             }
         }
         elsif ($tag) {
@@ -584,7 +646,7 @@ sub parse {
     if ($self->get_close_open_tags) {
         while (my $opened = pop @opened) {
             $self->_add_error('unclosed', $opened);
-            $self->_finish_tag($opened, '[/' . $opened->get_name . ']');
+            $self->_finish_tag($opened, '[/' . $opened->get_name . ']', 1);
             $callback_found_tag->($opened);
         }
     }
@@ -603,7 +665,7 @@ sub parse {
     }
     #warn __PACKAGE__.':'.__LINE__.": !!!!!!!!!!!! left text: '$text'\n";
     #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\@tags], ['tags']);
-    my $tree = Parse::BBCode::Tag->new({
+    my $tree = $self->new_tag({
         name => '',
         content => [@tags],
         start => '',
@@ -614,6 +676,11 @@ sub parse {
     return $tree;
 }
 
+sub new_tag {
+    my $self = shift;
+    Parse::BBCode::Tag->new(@_)
+}
+
 sub _add_error {
     my ($self, $error, $tag) = @_;
     my $errors = $self->get_error || {};
@@ -700,7 +767,7 @@ sub _render_tree {
                 not ref $_
             } @$fallback;
         }
-        if (($tree->get_class || '') eq 'block') {
+        if ($self->get_strip_linebreaks and ($tree->get_class || '') eq 'block') {
             if (@$content == 1 and not ref $content->[0] and defined $content->[0]) {
                 $content->[0] =~ s/^\r?\n//;
                 $content->[0] =~ s/\r?\n\z//;
@@ -749,46 +816,72 @@ sub escape_html {
     return $str;
 }
 
-sub _validate_attr {
-    my ($self, $tag, $attr) = @_;
-    $tag->set_attr_raw($attr);
-    my @array;
-    unless (length $attr) {
-        $tag->set_attr([]);
-        return 1;
+sub parse_attributes {
+    my ($self, %args) = @_;
+    my $text = $args{text};
+    my $tagname = $args{tag};
+    my $attribute_quote = $self->get_attribute_quote;
+    my $attr_string = '';
+    my $attributes = [];
+    if (
+        ($self->get_direct_attribute and $$text =~ s/^(=[^\]]*)?]//)
+            or
+        ($$text =~ s/^( [^\]]*)?\]//)
+    ) {
+        my $attr = $1;
+        my $end = ']';
+        $attr = '' unless defined $attr;
+        $attr_string = $attr;
+        unless (length $attr) {
+            return (1, [], $attr_string, $end);
+        }
+        if ($self->get_direct_attribute) {
+            $attr =~ s/^=//;
+        }
+        if ($self->get_strict_attributes and not length $attr) {
+            return (0, [], $attr_string, $end);
+        }
+        my @array;
+        if (length($attribute_quote) == 1) {
+            if ($attr =~ s/^(?:$attribute_quote(.+?)$attribute_quote|(.*?)(?:\s+|$))//) {
+                my $val = defined $1 ? $1 : $2;
+                push @array, [$val];
+            }
+            while ($attr =~ s/^([a-zA-Z0-9_]+)=(?:$attribute_quote(.+?)$attribute_quote|(.*?)(?:\s+|$))//) {
+                my $name = $1;
+                my $val = defined $2 ? $2 : $3;
+                push @array, [$name, $val];
+            }
+        }
+        else {
+            if ($attr =~ s/^(?:(["'])(.+?)\1|(.*?)(?:\s+|$))//) {
+                my $val = defined $2 ? $2 : $3;
+                push @array, [$val];
+            }
+            while ($attr =~ s/^([a-zA-Z0-9_]+)=(?:(["'])(.+?)\2|(.*?)(?:\s+|$))//) {
+                my $name = $1;
+                my $val = defined $3 ? $3 : $4;
+                push @array, [$name, $val];
+            }
+        }
+        if ($self->get_strict_attributes and length $attr and $attr =~ tr/ //c) {
+            return (0, [], $attr_string, $end);
+        }
+        $attributes = [@array];
+        return (1, $attributes, $attr_string, $end);
     }
-    if ($self->get_direct_attribute) {
-        $attr =~ s/^=//;
-    }
-    if ($self->get_strict_attributes and not length $attr) {
-        return 0;
-    }
-    if ($attr =~ s/^(?:"([^"]+)"|(.*?)(?:\s+|$))//) {
-        my $val = defined $1 ? $1 : $2;
-        push @array, [$val];
-    }
-    while ($attr =~ s/^([a-zA-Z0-9]+)=(?:"([^"]+)"|(.*?)(?:\s+|$))//) {
-        my $name = $1;
-        my $val = defined $2 ? $2 : $3;
-        push @array, [$name, $val];
-    }
-    #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\@array], ['array']);
-    #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$attr], ['attr']);
-    if ($self->get_strict_attributes and length $attr) {
-        return 0;
-    }
-    $tag->set_attr(\@array);
-    return 1;
+    return (0, $attributes, $attr_string, '');
 }
 
 # TODO add callbacks
 sub _finish_tag {
-    my ($self, $tag, $end) = @_;
+    my ($self, $tag, $end, $auto_closed) = @_;
     #warn __PACKAGE__.':'.__LINE__.": _finish_tag(@_)\n";
     #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$tag], ['tag']);
     unless ($tag->get_finished) {
         $tag->set_end($end);
         $tag->set_finished(1);
+        $tag->set_auto_closed($auto_closed || 0);
     }
     return 1;
 }
@@ -803,6 +896,18 @@ Parse::BBCode - Module to parse BBCode and render it as HTML or text
 
 =head1 SYNOPSIS
 
+Parse::BBCode parses common bbcode like
+
+    [b]bold[/b] [size=10]big[/size]
+
+short tags like
+
+    [foo://test]
+
+and custom bbcode tags.
+
+For the documentation of short tags, see L<"SHORT TAGS">.
+
 To parse a bbcode string, set up a parser with the default HTML defintions
 of L<Parse::BBCode::HTML>:
 
@@ -886,7 +991,7 @@ Constructor. Takes a hash reference with options as an argument.
             link => ...,
         },
         close_open_tags   => 1, # default 0
-        strict_attributes => 0, # default 0
+        strict_attributes => 0, # default 1
         direct_attributes => 1, # default 1
         url_finder        => 1, # default 0
         smileys           => 0, # default 0
@@ -954,10 +1059,14 @@ See L<"TEXT PROCESSORS">
 
 =item close_open_tags
 
+Default: 0
+
 If set to true (1), it will close open tags at the end or before block tags.
 
 =item strict_attributes
 
+Default: 1
+
 If set to true (1), tags with invalid attributes are left unparsed. If set to
 false (0), the attribute for this tags will be empty.
 
@@ -970,7 +1079,7 @@ you'd like to have this.
 
 =item direct_attributes
 
-Default: true
+Default: 1
 
 Normal tag syntax is:
 
@@ -980,6 +1089,45 @@ If set to 0, tag syntax is
 
   [tag attr2=val2 ...]
 
+=item attribute_quote
+
+You can change how the attribute values shuold be quoted.
+Default is a double quote (which is still optional):
+
+  my $parser = Parse::BBCode->new(
+      attribute_quote => '"',
+      ...
+  );
+  [tag="foo" attr="bar" attr2=baz]...[/tag]
+
+If you set it to single quote:
+
+  my $parser = Parse::BBCode->new(
+      attribute_quote => "'",
+      ...
+  );
+  [tag='foo' attr=bar attr2='baz']...[/tag]
+
+You can also set it to both: C<'">. Then both quoting types are
+allowed:
+
+  my $parser = Parse::BBCode->new(
+      attribute_quote => q/'"/,
+      ...
+  );
+  [tag='foo' attr="bar" attr2=baz]...[/tag]
+
+=item attribute_parser
+
+You can pass a subref that overrides the default attribute parsing.
+See L<"ATTRIBUTE PARSING">
+
+=item strip_linebreaks
+
+Default: 1
+
+Strips linebreaks at start/end of block tags
+
 =back
 
 =item render
@@ -1065,6 +1213,21 @@ the tree and return the raw text from it:
         my $corrected = $tree->raw_text;
     }
 
+=item parse_attributes
+
+You can inherit from Parse::BBCode and define your own attribute
+parsing. See L<"ATTRIBUTE PARSING">.
+
+=item new_tag
+
+Returns a L<Parse::BBCode::Tag> object.
+It just does:
+    shift;
+    Parse::BBCode::Tag->new(@_);
+
+If you want your own tag class, inherit from Parse::BBCode and
+let it return Parse::BBCode::YourTag->new
+
 =back
 
 
@@ -1373,6 +1536,69 @@ This will use the default like shown above (max length 50 chars).
 
 Default is 0.
 
+=head1 ATTRIBUTES
+
+There are two types of tags. The default (option direct_attributes=1):
+
+    [foo=bar a=b c=d]
+    [foo="text with space" a=b c=d]
+
+The parsed attribute structure will look like:
+
+    [ ['bar'], ['a' => 'b'], ['c' => 'd'] ]
+
+Another bbcode variant doesn't use direct attributes:
+
+    [foo a=b c=d]
+
+The resulting attribute structure will have an empty first element:
+
+    [ [''], ['a' => 'b'], ['c' => 'd'] ]
+
+=head1 ATTRIBUTE PARSING
+
+If you have bbcode attributes that don't fit into the two standard
+syntaxes you can inherit from Parse::BBCode and overwrite the
+parse_attributes method, or you can pass an option attribute_parser
+contaning a subref.
+
+Example:
+
+    [size=10]big[/size] [foo|bar|boo]footext[/foo] end
+
+The size tag should be parsed normally, the foo tag needs different parsing.
+
+    sub parse_attributes {
+        my ($self, %args) = @_;
+        # $$text contains '|bar|boo]footext[/foo] end
+        my $text = $args{text};
+        my $tagname = $args{tag}; # 'foo'
+        if ($tagname eq 'foo') {
+            # work on $$text
+            # result should be something like:
+            # $$text should contain 'footext[/foo] end'
+            $valid = 1;
+            @attr = ( [''], [1 => 'bar'], [2 => 'boo'] );
+            $attr_string = '|bar|boo';
+            return ($valid, [@attr], $attr_string, ']');
+        }
+        else {
+            return shift->SUPER::parse_attributes(@_);
+        }
+    }
+    my $parser = Parse::BBCode->new({
+        ...
+        attribute_parser => \&parse_attributes,
+    });
+
+If the attributes are not valid, return
+
+    0, [ [''] ], '|bar|boo', ']'
+
+If you don't find a closing square bracket, return:
+
+    0, [ [''] ], '|bar|boo', ''
+
 =head1 TEXT PROCESSORS
 
 If you set url_finder and linebreaks to 1, the default text processor
@@ -1435,6 +1661,62 @@ special tag with the empty string:
             },
             ...
 
+=head1 SHORT TAGS
+
+It can be very convenient to have short tags like [foo://id].
+This is not really a part of BBCode, but I consider it as quite similar,
+so I added it to this module.
+For example to link to threads, cpan modules or wikipedia articles:
+
+    [thread://123]
+    [thread://123|custom title]
+    # can be implemented so that it links to thread 123 in the forum
+    # and additionally fetch the thread title.
+
+    [cpan://Module::Foo|some useful module]
+
+    [wikipedia://Harold & Maude]
+
+You can define a short tag by adding the option C<short>. The tag will
+work as a classic tag and short tag. If you only want to support the
+short version, set the option C<classic> to 0.
+
+    my $p = Parse::BBCode->new({
+            tags => {
+                Parse::BBCode::HTML->defaults,
+                wikipedia => {
+                    short   => 1,
+                    output  => '<a href="http://wikipedia/?search=%{uri}A">%{parse}s</a>',
+                    class   => 'url',
+                    classic => 0, # don't support classic [wikipedia]...[/wikipedia]
+                },
+                thread => {
+                    code => sub {
+                        my ($parser, $attr, $content, $attribute_fallback) = @_;
+                        my $id = $attribute_fallback;
+                        if ($id =~ tr/0-9//c) {
+                            return '[thread]' . encode_entities($id) . '[/thread]';
+                        }
+                        my $name;
+                        if ($attr) {
+                            # custom title will be in $attr
+                            # [thread=123]custom title[/thread]
+                            # [thread://123|custom title]
+                            # already escaped
+                            $name = $$content;
+                        }
+                        return qq{<a href="/thread/$id">$name</a>};
+                    },
+                    short   => 1,
+                    classic => 1, # default is 1
+                },
+            },
+        }
+    );
+
+
+
+
 =head1 WHY ANOTHER BBCODE PARSER
 
 I wrote this module because L<HTML::BBCode> is not extendable (or
@@ -1457,6 +1739,36 @@ a tag for linking to perlmonks.org, other forums need other tags.
 Another goal was to always output a result and don't die. I might add an
 option which lets the parser die with unbalanced code.
 
+=head1 WHY BBCODE?
+
+Some forums and blogs prefer a kind of pseudo HTML for user comments.
+The arguments against bbcode is usually: "Why should people learn an
+additional markup language if they can just use HTML?" The problem is
+that many people don't know HTML.
+
+BBCode is often a bit shorter, for example if you have a code tag
+with an attribute that tells the parser what language the content is in.
+
+    [code=perl]...[/code]
+    <code language="perl">...</code>
+
+Also, forum HTML is usually not real HTML. It is usually a subset and
+sometimes with additional tags.
+So in the backend you need to parse it anyway to turn it into real HTML.
+
+BBCode is widely known and used.
+Unfortunately though, there is no specification; some forums only allow
+attributes in double quotes, some forums implement only one attribute that
+can be seperated by spaces, which makes it difficult to parse if you
+want to support more than one attribute.
+
+I tried to support the most common syntax (attributes without quotes, in
+single or double quotes) and tags.
+If you need additional tags it's relatively easy to implement them.
+For example in my forum I implemented a [more] tag that hides long
+text or code in thread view. Without Javascript you will see the expanded
+content when clicking on the single article, or with Javascript the
+content will be added inline via Ajax.
 
 =head1 TODO
 
@@ -1471,29 +1783,7 @@ roughly tested.
 
 The main syntax is likely to stay, only the API for callbacks
 might change. At the moment it is not possible to add callbacks
-to the parsing process, only for the rendering phase. It is
-also not possible to declare your own attribute syntax, for
-example
-
-    [quote=nickname date]
-
-Attributes always have to look like:
-
-    [tag=main_attribute other=foo]...
-    [tag="main_attribute" other="foo"]...
-
-I would like to add support for different syntax, because it might
-happen that one has old bbcode lying around (maybe taken over from
-a different forum software) and cannot manually replace all the invalid
-bbcode.
-
-=item Short bbcode like syntax for urls
-
-Since I need this myself in my board software and currently do the
-replacing in the text processor I would like to add tags like this:
-
-    [cpan://This::Module|title]
-
+to the parsing process, only for the rendering phase.
 
 =back
 
